/*
 * Copyright (C) 2004 ruby-ecore team (see AUTHORS)
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <ruby.h>
#include <stdbool.h>

#include <Ecore.h>

#define __RB_EVENT_HANDLER_C
#include "rb_ecore.h"
#include "rb_event_handler.h"

typedef struct {
	Ecore_Event_Handler *real;
	VALUE event_class;
	VALUE callback;
	bool deleted;
} RbEventHandler;

static int on_ecore_event (void *data, int type, void *event);
static VALUE c_ev_raise (VALUE klass, VALUE argv);

VALUE event_classes, cEcoreEvent;
static VALUE event_args;

static void c_mark (RbEventHandler *h)
{
	rb_gc_mark (h->callback);
}

static void c_free (RbEventHandler *h)
{
	if (h->real && !h->deleted)
		ecore_event_handler_del (h->real);

	free (h);
}

static VALUE c_alloc (VALUE klass)
{
	RbEventHandler *h = NULL;

	return Data_Make_Struct (klass, RbEventHandler, c_mark, c_free, h);
}

static VALUE c_init (VALUE self, VALUE klass)
{
	RbEventHandler *h = NULL;
	int t;

	if (!rb_block_given_p ())
		rb_raise (rb_eStandardError, "block missing");

	Data_Get_Struct (self, RbEventHandler, h);

	if (rb_obj_is_kind_of (klass, rb_cModule) != Qtrue)
		rb_raise (rb_eArgError, "invalid argument");

	t = NUM2INT (rb_const_get (klass, rb_intern ("TYPE")));
	if (t <= ECORE_EVENT_NONE)
		rb_raise (rb_eStandardError, "invalid event");

	h->event_class = klass;
	h->callback = rb_block_proc ();
	h->deleted = false;
	h->real = ecore_event_handler_add (t, on_ecore_event, h);

	return self;
}

static VALUE c_delete (VALUE self)
{
	RbEventHandler *h = NULL;

	Data_Get_Struct (self, RbEventHandler, h);

	if (h->real && !h->deleted) {
		ecore_event_handler_del (h->real);
		h->real = NULL;
		h->deleted = true;
	} else
		rb_raise (rb_eException, "EventHandler already deleted!");

	return Qnil;
}

static int on_ecore_event (void *data, int type, void *event)
{
	RbEventHandler *h = data;
	VALUE obj, tmp = (VALUE) event;

	obj = rb_obj_alloc (h->event_class);

	if (rb_respond_to (h->event_class, rb_intern ("raise")))
		rb_apply (obj, rb_intern ("initialize"), tmp);
	else
		rb_obj_call_init (obj, 1, &tmp);

	tmp = rb_funcall (h->callback, rb_intern ("call"), 1, obj);

	/* if the block returned false, don't call the other
	 * event handlers
	 */
	return (tmp != Qfalse);
}

VALUE c_ev_inherited (VALUE klass, VALUE child)
{
	VALUE t;

	t = INT2FIX (ecore_event_type_new ());
	rb_hash_aset (event_classes, t, child);

	rb_define_const (child, "TYPE", t);
	rb_define_singleton_method (child, "raise", c_ev_raise, -2);

	return Qnil;
}

static VALUE c_ev_init (int argc, VALUE *argv, VALUE self)
{
	return self;
}

static void free_event_args (void *data, void *argv)
{
	rb_ary_delete (event_args, (VALUE) argv);
}

static VALUE c_ev_raise (VALUE klass, VALUE argv)
{
	VALUE t;

	rb_ary_push (event_args, argv);

	t = rb_const_get (klass, rb_intern ("TYPE"));
	ecore_event_add (FIX2INT (t), (void *) argv, free_event_args, NULL);

	return Qnil;
}

void Init_EventHandler (void)
{
	VALUE cEventHandler;

	cEventHandler = rb_define_class_under (mEcore, "EventHandler",
	                                       rb_cObject);

	rb_define_alloc_func (cEventHandler, c_alloc);
	rb_define_method (cEventHandler, "initialize", c_init, 1);
	rb_define_method (cEventHandler, "delete", c_delete, 0);

	event_classes = rb_hash_new ();
	rb_global_variable (&event_classes);

	event_args = rb_ary_new ();
	rb_global_variable (&event_args);

	/* define a base event class */
	cEcoreEvent = rb_define_class_under (mEcore, "Event", rb_cObject);
	rb_define_singleton_method (cEcoreEvent, "inherited", c_ev_inherited, 1);
	rb_define_method (cEcoreEvent, "initialize", c_ev_init, -1);
}
