/*
 * Copyright (C) 2004 ruby-ecore team (see AUTHORS)
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <ruby.h>

#include <Ecore.h>
#include <Ecore_Job.h>
#include <stdbool.h>

#include "rb_ecore_job.h"

typedef struct {
	Ecore_Job *real;
	VALUE callback;
	bool deleted;
} RbJob;

static void on_job (void *data)
{
	RbJob *job = data;

	rb_funcall (job->callback, rb_intern ("call"), 0);
	job->deleted = true;
}

static void c_mark (RbJob *job)
{
	rb_gc_mark (job->callback);
}

static void c_free (RbJob *job)
{
	if (job->real && !job->deleted)
		ecore_job_del (job->real);

	ecore_shutdown ();

	free (job);
}

static VALUE c_alloc (VALUE klass)
{
	RbJob *job = NULL;

	ecore_init ();

	return Data_Make_Struct (klass, RbJob, c_mark, c_free, job);
}

/*
 * call-seq:
 *  Ecore::Job::Job.new { block } => job
 *
 * Creates an Ecore::Job::Job object.
 * After execution, the object will be deleted.
 */
static VALUE c_init (VALUE self)
{
	RbJob *job = NULL;

	if (!rb_block_given_p ())
		rb_raise (rb_eStandardError, "block missing");

	Data_Get_Struct (self, RbJob, job);

	job->callback = rb_block_proc ();
	job->deleted = false;
	job->real = ecore_job_add (on_job, job);

	return self;
}

/*
 * call-seq:
 *  job.delete => nil
 *
 * Deletes <i>job</i>.
 */
static VALUE c_delete (VALUE self)
{
	VALUE ret = Qfalse;
	RbJob *job = NULL;

	Data_Get_Struct (self, RbJob, job);

	if (job->real && !job->deleted) {
		ecore_job_del (job->real);
		job->deleted = true;
		job->real = NULL;
		ret = Qtrue;
	}

	return ret;
}

void Init_Job (void)
{
	VALUE c = rb_define_class_under (mJob, "Job", rb_cObject);

	rb_define_alloc_func (c, c_alloc);
	rb_define_method (c, "initialize", c_init, 0);
	rb_define_method (c, "delete", c_delete, 0);
}
